from django.db import models
from django.contrib.contenttypes import generic
from django.contrib.contenttypes.models import ContentType

from .fields import RoleField, PermissionField

class RolePermission(models.Model):
    permission = PermissionField(max_length=255, blank=False, db_index=True)
    role = RoleField(max_length=255, blank=False, db_index=True)

    content_type = models.ForeignKey(ContentType)
    object_id = models.PositiveIntegerField(db_index=True)
    content_object = generic.GenericForeignKey('content_type', 'object_id')

    @classmethod
    def clear(cls, obj, permission):
        """ clear all role assignments for permission on instance """
        model_ct = ContentType.objects.get_for_model(obj)
        cls.objects.filter(content_type=model_ct, object_id=obj.id,
                           permission=permission).delete()

    @classmethod
    def assign(cls, obj, role, permission):
        model_ct = ContentType.objects.get_for_model(obj)
        r, _ = cls.objects.get_or_create(content_type=model_ct,
                                         object_id=obj.id,
                                         permission=permission,
                                         role=role)
        return r

    @classmethod
    def assignments(cls, obj):  # XXX make this a manager?
        """ return all assignments for a specific object """
        model_ct = ContentType.objects.get_for_model(obj)
        return cls.objects.filter(content_type=model_ct, object_id=obj.id).all()


    def __unicode__(self):
        return u"<Permission {0} for role {1}>".format(self.permission,
                                                      self.role)



from django.contrib.contenttypes.models import ContentType


class Base(object):
    """
        A singletonish (based on id) type with additional
        equality functionality
    """
    _registry = None  ## should be overridden, don't share through base

    def __init__(self, id, name="", description=""):
        self.id = id
        self.name = name or id
        self.description = description

    def __eq__(self, other):
        return other and self.__class__ is other.__class__ and \
               self.id == other.id

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    @classmethod
    def create(cls, id, name="", description=""):
        p = cls._registry.get(id)
        if not p:
            p = cls(id, name, description)
            cls._registry[id] = p
        return p

    @classmethod
    def all(cls):
        return cls._registry.values()

class Permission(Base):
    _registry = {}

    def __unicode__(self):
        return u"<Permission {0} ({1})>".format(self.id, self.name)

    __repr__ = __unicode__

class Role(Base):
    _registry = {}

    def has_access(self, obj, permission):
        from .models import RolePermission
        model_ct = ContentType.objects.get_for_model(obj)
        return RolePermission.objects.filter(content_type=model_ct,
                                             object_id=obj.id,
                                             role=self,
                                             permission=permission).exists()

    def __unicode__(self):
        return u"<Role {0} ({1})>".format(self.id, self.name)

    __repr__ = __unicode__



import pytest

from twotest.fixtures import client, django_client

from drole.types import Role, Permission
from drole.models import RolePermission
from drole.test_models import TestModel

@pytest.fixture()
def role():
    return Role("test.arole", "A role")

@pytest.fixture()
def anonymous_role():
    return Role("test.anonymous", "Anonymous")

@pytest.fixture()
def permission():
    return Permission("test.permission", "Test Permission")

@pytest.fixture()
def view_permission():
    return Permission("test.view", "View Permission")

class TestAuthentication(object):
    def test_unicode(self, client, role, permission):
        obj = TestModel.objects.get_or_create(name="bla")[0]
        assert isinstance(RolePermission.assign(obj,
                                                role,
                                                permission).__unicode__(),
                          unicode)

    def test_empty(self, client, role, permission):
        """ no assignments at all """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        assert not role.has_access(obj, permission)

    def test_mismatch(self, client, anonymous_role, role, view_permission):
        """ permission assigned to different role """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, role, view_permission)

        assert not anonymous_role.has_access(obj, view_permission)

    def test_cross_mismatch(self, client, anonymous_role, role,
                            view_permission, permission):
        """ cross assignment of two perms and two roles """
        obj1 = TestModel.objects.get_or_create(name="1")[0]
        obj2 = TestModel.objects.get_or_create(name="2")[0]

        RolePermission.assign(obj1, role, permission)
        RolePermission.assign(obj2, anonymous_role, view_permission)

        assert not anonymous_role.has_access(obj1, view_permission)
        assert not anonymous_role.has_access(obj2, permission)
        assert not role.has_access(obj1, view_permission)
        assert not role.has_access(obj2, permission)

    def test_cross_match(self, client, anonymous_role, role,
                         view_permission, permission):
        """ cross assignment of two perms and two roles """
        obj1 = TestModel.objects.get_or_create(name="1")[0]
        obj2 = TestModel.objects.get_or_create(name="2")[0]

        RolePermission.assign(obj1, role, permission)
        RolePermission.assign(obj2, anonymous_role, view_permission)

        assert anonymous_role.has_access(obj2, view_permission)
        assert role.has_access(obj1, permission)

    def test_has_access(self, client, anonymous_role, view_permission):
        """ simple success case: role/perm assignment """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, anonymous_role, view_permission)
        assert anonymous_role.has_access(obj, view_permission)

    def test_assignments_empty(self, client):
        """ by default no assignments """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        assert not RolePermission.assignments(obj).exists()

    def test_assignments(self, client, anonymous_role, view_permission):
        """ simple success case: role/perm assignment """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, anonymous_role, view_permission)
        assert RolePermission.assignments(obj).exists()

    def test_assignments_dup(self, client, anonymous_role, view_permission):
        """ a duplicate assignment should not create duplicate records """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, anonymous_role, view_permission)
        RolePermission.assign(obj, anonymous_role, view_permission)
        assert RolePermission.assignments(obj).count() == 1

    def test_clear_noeffect(self, client, anonymous_role):
        """ clear a permission that has no assignment """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, anonymous_role, Permission("one"))

        RolePermission.clear(obj, Permission("other"))
        assert RolePermission.assignments(obj).count() == 1

    def test_clear(self, client, anonymous_role):
        """ clear a permission that has an assignment """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, anonymous_role, Permission("one"))

        RolePermission.clear(obj, Permission("one"))
        assert RolePermission.assignments(obj).count() == 0

    def test_clear_complex(self, client):
        """ clear a permission that has an multiple assignments """
        obj = TestModel.objects.get_or_create(name="bla")[0]

        RolePermission.assign(obj, Role("r1"), Permission("one"))
        RolePermission.assign(obj, Role("r2"), Permission("one"))
        RolePermission.assign(obj, Role("r1"), Permission("two"))
        RolePermission.assign(obj, Role("r2"), Permission("two"))

        RolePermission.clear(obj, Permission("one"))
        a =  RolePermission.assignments(obj)

        assert a.count() == 2
        assert set(e.role for e in a.all()) == set((Role("r1"), Role("r2")))
        assert set(e.permission for e in a.all()) == set((Permission("two"),))




from django.db import models
from .types import Role, Permission

class RoleField(models.CharField):

    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        if value is None:
            return None

        if isinstance(value, (str, unicode)):
            return Role(value)

        return value

    def get_prep_value(self, value):
        if isinstance(value, Role):
            return value.id
        return value


class PermissionField(models.CharField):

    __metaclass__ = models.SubfieldBase

    def to_python(self, value):
        if value is None:
            return None

        if isinstance(value, (str, unicode)):
            return Permission(value)

        return value

    def get_prep_value(self, value):
        if isinstance(value, Permission):
            return value.id
        return value

from south.modelsinspector import add_introspection_rules
add_introspection_rules([], ["^drole\.fields\.RoleField"])
add_introspection_rules([], ["^drole\.fields\.PermissionField"])



import pytest
from drole.types import Permission, Role

@pytest.fixture
def clean_registry():
    Permission._registry.clear()

@pytest.mark.usefixtures("clean_registry")    
class TestPermission(object):
    """
        Permission is a simple wrapper around a string identifier
        with some singleton-ish behaviour
    """
    def test_unicode(self):
        assert isinstance(Permission.create("foo").__unicode__(), unicode)

    def test_create(self):
        """ Permissions are implicitly created """
        assert Permission.create("foo") is not None

    def test_uniqueness_id(self):
        """ Different instances with same identifier are identical """
        assert Permission.create("foo") is Permission.create("foo")

    def test_uniqueness_attrs(self):
        """ Different instances with same identifier are identical, even
            if other attributes differ """
        assert Permission.create("foo", "bar", "bla") is \
               Permission.create("foo", "this", "that")

    def test_uniqueness_differ(self):
        """ 
            Permissions with different identifier are not equal
        """
        assert Permission.create("foo") != Permission.create("bar")

    def test_equality(self):
        """ ordinary comparison. Will succeed since it will be the same
            objects """
        assert Permission.create("foo") == Permission.create("foo")

    def test_forced_equality(self):
        """ if somehow different permissions get created with the same
            identifier, they should still be equal """
        assert Permission("foo") == Permission("foo")

    def test_forced_inequality(self):
        """ if somehow different permissions get created with different
            identifiers, they shouldn't be equal """
        assert Permission("foo") != Permission("bar")

    def test_forced_identity(self):
        """ if somehow different permissions get created with the same
            identifier, they are equal but not identical """
        assert Permission("foo") is not Permission("foo")

    def test_in(self):
        """ a common case """
        assert Permission("foo") in [Permission("bar"),
                                     Permission("foo"), Permission("hello")]

    def test_all_empty(self):
        """ empty role collection """
        assert len(Permission.all()) == 0

    def test_all(self):
        """ retrieve all created permissions """
        p1 = Permission.create("r1")
        p2 = Permission.create("r2")
        p3 = Permission.create("r2") # duplicate!
        assert set(Permission.all()) == set((p1, p2))

    def test_hash(self):
        """ verify permissions can be hashed properly """
        assert set((Permission("r1"),)) == set((Permission("r1"),))

    def test_hash_perm_role(self):
        """ Permission / Role are hashed differently, even with same id """
        assert set((Permission("r1"),)) != set((Role("r1"),))



from drole.fields import RoleField, PermissionField
from drole.types import Role, Permission

class BaseFieldTest(object):
    field = None
    type = None

    def test_none_to_python(self):
        f = self.field()
        assert f.to_python(None) is None

    def test_string_to_python(self):
        f = self.field()
        assert f.to_python('x') == self.type('x')

    def test_type_to_python(self):
        f = self.field()
        assert f.to_python(self.type('x')) == self.type('x')

    def test_string_to_db(self):
        f = self.field()
        assert f.get_prep_value('string') == 'string'

class TestRoleField(BaseFieldTest):
    field = RoleField
    type = Role

class TestPermissionField(BaseFieldTest):
    field = PermissionField
    type = Permission




import pytest
from drole.types import Role

@pytest.fixture
def clean_registry():
    Role._registry.clear()

@pytest.mark.usefixtures("clean_registry")    
class TestRole(object):
    """
        Role is a simple wrapper around a string identifier
        with some singleton-ish behaviour
    """
    def test_unicode(self):
        assert isinstance(Role.create("foo").__unicode__(), unicode)

    def test_create(self):
        """ Role are implicitly created """
        assert Role.create("foo") is not None

    def test_uniqueness_id(self):
        """ Different instances with same identifier are identical """
        assert Role.create("foo") is Role.create("foo")

    def test_uniqueness_attrs(self):
        """ Different instances with same identifier are identical, even
            if other attributes differ """
        assert Role.create("foo", "bar", "bla") is \
               Role.create("foo", "this", "that")

    def test_uniqueness_differ(self):
        """ 
            Role with different identifier are not equal
        """
        assert Role.create("foo") != Role("bar")

    def test_equality(self):
        """ ordinary comparison. Will succeed since it will be the same
            objects """
        assert Role.create("foo") == Role.create("foo")

    def test_forced_equality(self):
        """ if somehow different roles get created with the same identifier,
            they should still be equal """
        assert Role("foo") == Role("foo")

    def test_forced_inequality(self):
        """ if somehow different roles get created with different identifiers,
            they should not be equal """
        assert Role("foo") != Role("bar")

    def test_forced_identity(self):
        """ if somehow different roles get created with the same identifier,
            they are equal but not identical """
        assert Role("foo") is not Role("foo")

    def test_in(self):
        """ a common case """
        assert Role("foo") in [Role("bar"), Role("foo"), Role("hello")]

    def test_all_empty(self):
        """ empty role collection """
        assert len(Role.all()) == 0

    def test_all(self):
        """ retrieve all created permissions """
        r1 = Role.create("r1")
        r2 = Role.create("r2")
        r3 = Role.create("r2") # duplicate!
        assert set(Role.all()) == set((r1, r2))

    def test_hash(self):
        """ verify roles can be hashed properly """
        assert set((Role("r1"),)) == set((Role("r1"),))




from setuptools import setup, find_packages
import os

version = '0.1.0'

setup(name='django_drole',
      version=version,
      description="Django Drole: Zope/Plone inspired roles/permissions",
      long_description=open("README.md").read(),
      classifiers=[
        "Programming Language :: Python",
        ],
      keywords='',
      author='Ivo van der Wijk',
      author_email='two@in.m3r.nl',
      url='http://github.com/iivvoo/django_drole',
      license='BSD',
      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=[],
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'setuptools',
          'pytest',
          'coverage',
      ],
      entry_points={
      },

      )

